from io import BytesIO
from pathlib import Path

# relative imports only! otherwise the module is not ofund in the published package
from .utils.merger import merge_xml
from .utils.xml import extract_document_xml, replace_document_xml


def read_file(file_path: str) -> BytesIO:
    with Path(file_path).open("rb") as file:
        stream = BytesIO(file.read())
        stream.seek(0)
    return stream


def merge_docx(
    source_path: str,
    content_path: str,
    output_path: str | None = None,
    pattern: str | None = None,
    insert_start: bool = False,
    insert_end: bool = False,
) -> BytesIO:
    if not source_path:
        raise ValueError("Source file path cannot be empty.")

    if not content_path:
        raise ValueError("Content file path cannot be empty.")

    if not isinstance(source_path, str) or not isinstance(content_path, str):
        raise TypeError("Source and content file paths must be strings.")

    if not isinstance(output_path, (str, type(None))):
        raise TypeError("Output file path must be a string or None.")

    if not isinstance(pattern, (str, type(None))):
        raise TypeError("Pattern must be a string or None.")

    if not isinstance(insert_start, bool):
        raise TypeError("insert_start must be a boolean.")

    if not isinstance(insert_end, bool):
        raise TypeError("insert_end must be a boolean.")

    if not source_path.endswith(".docx"):
        raise ValueError("Source file must be a DOCX file.")

    if not content_path.endswith(".docx"):
        raise ValueError("Content file must be a DOCX file.")

    if output_path and not output_path.endswith(".docx"):
        raise ValueError("Output file must be a DOCX file.")

    if not Path(source_path).is_file():
        raise FileNotFoundError(f"Source file {source_path} does not exist.")

    if not Path(content_path).is_file():
        raise FileNotFoundError(f"Content file {content_path} does not exist.")

    if output_path and Path(output_path).is_file():
        raise FileExistsError(f"Output file {output_path} already exists.")

    if pattern is None and not insert_start and not insert_end:
        raise ValueError("At least one of pattern, insert_start, or insert_end must be specified.")

    source_document_stream = extract_document_xml(source_path)
    content_document_stream = extract_document_xml(content_path)

    # Create the final document.xml stream
    output_document_xml_stream = merge_xml(
        source_xml_stream=source_document_stream,
        content_xml_stream=content_document_stream,
        pattern=pattern,
        insert_start=insert_start,
        insert_end=insert_end,
    )

    # Replace the document.xml in the template DOCX with the final document.xml
    template_stream = read_file(content_path)
    output_stream = replace_document_xml(template_stream, output_document_xml_stream)

    if output_path:
        # Save the modified DOCX to a new file
        with Path(output_path).open("wb") as output_file:
            output_file.write(output_stream.getvalue())

    return output_stream



from copy import deepcopy
from io import BytesIO

# relative imports only! otherwise the module is not ofund in the published package
from .xml import get_string_from_xml, get_xml_from_stream


def merge_xml(
    source_xml_stream: BytesIO, content_xml_stream: BytesIO, pattern: str | None, insert_start: bool, insert_end: bool
) -> BytesIO:
    # Parse the content XML
    source_root = get_xml_from_stream(content_xml_stream)

    # Define the namespaces
    namespaces = {"w": "http://schemas.openxmlformats.org/wordprocessingml/2006/main"}

    # Find the <w:body> tag in the source XML
    source_body = source_root.find(".//w:body", namespaces)
    if source_body is None:
        raise ValueError("No <body> tag found in the content XML")

    # Filter out <w:sectPr> tags from the source XML
    filtered_elements = []
    for elem in source_body:
        if elem.tag == "{http://schemas.openxmlformats.org/wordprocessingml/2006/main}sectPr":
            continue
        filtered_elements.append(elem)

    # Parse the target template XML
    target_root = get_xml_from_stream(source_xml_stream)

    # Find the <w:body> tag in the target template XML
    target_body = target_root.find(".//w:body", namespaces)
    if target_body is None:
        raise ValueError("No <body> tag found in the target template XML")

    # Find the <w:sectPr> tag in the target XML
    sect_pr = target_body.find("w:sectPr", namespaces)
    if sect_pr is None:
        raise ValueError("No <sectPr> tag found in the target XML")

    index_offset = 0
    for index, elem in enumerate(target_body):
        text_elem = elem.find(".//w:t", namespaces)
        text_content = text_elem.text if text_elem is not None else ""
        if text_content == pattern:
            target_body.remove(elem)
            for content_elem_index, elem in enumerate(deepcopy(filtered_elements)):
                target_body.insert(index + index_offset + content_elem_index, elem)
            index_offset += len(filtered_elements)
        if elem.tag == "{http://schemas.openxmlformats.org/wordprocessingml/2006/main}sectPr":
            if insert_end:
                # Insert the filtered elements at the end of the body
                for content_elem_index, elem in enumerate(deepcopy(filtered_elements)):
                    target_body.insert(index + index_offset + content_elem_index, elem)
            break

    if insert_start:
        # Insert the filtered elements at the start of the body
        for content_elem_index, elem in enumerate(deepcopy(filtered_elements)):
            target_body.insert(content_elem_index, elem)

    if index_offset == 0 and not insert_start and not insert_end:
        # No pattern found and no insert_start or insert_end specified
        raise ValueError("No pattern found in the target XML and no insert_start or insert_end specified.")

    output_stream = BytesIO()
    output_stream.write(get_string_from_xml(target_root))
    output_stream.seek(0)
    return output_stream



import zipfile
from io import BytesIO

from lxml import etree


def extract_document_xml(file_path: BytesIO | str) -> BytesIO:
    with zipfile.ZipFile(file_path, "r") as zip_ref:
        with zip_ref.open("word/document.xml") as xml_file:
            document_xml_stream = BytesIO(xml_file.read())
            document_xml_stream.seek(0)
    return document_xml_stream


def replace_document_xml(docx_stream: BytesIO, document_xml_stream: BytesIO) -> BytesIO:
    output_zip_stream = BytesIO()
    with zipfile.ZipFile(docx_stream, "r") as docx_zip:
        with zipfile.ZipFile(output_zip_stream, "w", zipfile.ZIP_DEFLATED) as output_zip:
            for item in docx_zip.infolist():
                if item.filename != "word/document.xml":
                    output_zip.writestr(item, docx_zip.read(item.filename))
            output_zip.writestr("word/document.xml", document_xml_stream.getvalue())
    output_zip_stream.seek(0)
    return output_zip_stream


def get_xml_from_stream(file_stream: BytesIO):
    source_tree = etree.parse(file_stream)
    return source_tree.getroot()


def get_string_from_xml(xml_tree: etree._Element):
    # Convert the XML tree to a string
    return etree.tostring(xml_tree, encoding="UTF-8", xml_declaration=True, standalone=True)
